/*
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this code; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  */

/* Copyright (C) 2000 Ian Zimmerman <itz@speakeasy.org> */

static const char rcsid[] = "$Id: linuxonly.c,v 1.20 2002/09/12 03:37:46 itz Exp $";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <fcntl.h>
#include <paths.h>
#include <syslog.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/vt.h>
#include <sys/kd.h>
#include <termios.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <sys/types.h>
#include <time.h>

#include "linuxonly.h"
#include "readproc.h"
#include "gim_common.h"
#include "exc.h"
#include "bitsets.h"

const char devfs_format[] = _PATH_DEV "vc/%d";
const char nodevfs_format[] = _PATH_DEV "tty%d";

static int console_fd[MAX_NR_CONSOLES+1];
declare_bits(console_opened, MAX_NR_CONSOLES+1);
declare_bits(paste_disabled, MAX_NR_CONSOLES+1);
static time_t selection_time;

static struct winsize win[MAX_NR_CONSOLES+1];

int max_selection_age = -1;

#define TIOCLINUX_KBDMOD_MAGIC 6

static void
do_open_console (int vt)
{
  char console_path[50];
  int len;

  if (!test_bit(console_opened, vt)) {
    len = snprintf(console_path, sizeof(console_path), devfs_format, vt);
    assert(len >= 0 && len < sizeof(console_path));
    console_fd[vt] = open(console_path, O_RDONLY|O_NOCTTY);
    if (console_fd[vt] == -1) {
      len = snprintf(console_path, sizeof(console_path), nodevfs_format, vt);
      assert(len >= 0 && len < sizeof(console_path));
      console_fd[vt] = open(console_path, O_RDONLY|O_NOCTTY);
    } /*if*/
    if (console_fd[vt] == -1) {
      int my_errno = errno;
      syslog(LOG_ERR, "open(%s) failed: %m", console_path);
      exc_eraise2(unix_error, my_errno);
    } /*if*/
    set_bit(console_opened, vt);
  } /*if*/
}

static void
reopen_console (int vt)
{
  if (test_bit(console_opened, vt)) {
    clear_bit(console_opened, vt);
    close(console_fd[vt]);
  } /*if*/

  do_open_console(vt);
}

int
current_vt (void)
{
  struct vt_stat vts;

  do_open_console(0);    

  if (ioctl(console_fd[0], VT_GETSTATE, &vts) < 0) {
    int my_errno = errno;
    syslog(LOG_ERR, "ioctl(VT_GETSTATE) failed: %m");
    exc_eraise2(unix_error, my_errno);
  } /*if*/

  return vts.v_active;
}

int
is_text_mode (void)
{
  int vt_mode;
  int current;
  int tried;
  int vt;

  vt = current_vt();
  do_open_console(vt);

  for (tried = 0; tried < 2; tried++) {
    if (ioctl(console_fd[vt], KDGETMODE, &vt_mode) < 0) {
      int my_errno = errno;
      if (my_errno == EIO && tried == 0) {
        reopen_console(vt);
        continue;               /* back to for */
      } else {
        syslog(LOG_ERR, "ioctl(KDGETMODE) failed: %m");
        exc_eraise2(unix_error, my_errno);
      } /*if*/
    } /*if*/
    break;
  } /*for*/

  return vt_mode == KD_TEXT;
}

unsigned
get_kbd_modifiers (void)
{
  unsigned u;
  int tried;

  do_open_console(0);

  u = TIOCLINUX_KBDMOD_MAGIC;

  for (tried = 0; tried < 2; tried++) {
    if (ioctl(console_fd[0], TIOCLINUX, &u) < 0) {
      int my_errno = errno;
      if (my_errno == EIO && tried == 0) {
        reopen_console(0);
        continue;               /* back to for */
      } else {
        syslog(LOG_ERR, "ioctl(TIOCLINUX) failed: %m");
        exc_eraise2(unix_error, my_errno);
      } /*if*/
    } /*if*/
    break;
  } /*for*/
  
  return u;
}

void
get_window_size (int* px, int* py, int vt)
{
  int current;
  int tried;

  if (win[vt].ws_col == 0 || win[vt].ws_row == 0) {
    do_open_console(vt);
    for (tried = 0; tried < 2; tried++) {
      if (ioctl(console_fd[vt], TIOCGWINSZ, &win[vt]) < 0) {
        int my_errno = errno;
        if (my_errno == EIO && tried == 0) {
          reopen_console(vt);
          continue;               /* back to for */
        } else {
          syslog(LOG_ERR, "ioctl(TIOCGWINSZ) failed: %m");
          exc_eraise2(unix_error, my_errno);
        } /*if*/
      } /*if*/
      break;
    } /*for*/
  } /*if*/

  *px = win[vt].ws_col;
  *py = win[vt].ws_row;
}

int
is_stopped (pid_t pid)
{
  struct proc_s ps;

  if (look_up_pid(pid, &ps) < 0) {
    int my_errno = errno;
    syslog(LOG_WARNING, "cannot read proc information for %d: %m", pid);
    if (my_errno != ENOENT) {
      exc_eraise2(unix_error, my_errno);
    } else {
      return 1;
    } /*if*/
  } /*if*/

  return (ps.state == 'T');
}

void
draw_pointer (int x, int y, int vt)
{
  union {
    unsigned short ushort[6];
    unsigned char uchr[12];
  } u;
  int tried;

  do_open_console(vt);
  
  for (tried = 0; tried < 2; tried++) {
    u.uchr[1] = 2;
    u.ushort[1] = u.ushort[3] = (unsigned short)(x+1);
    u.ushort[2] = u.ushort[4] = (unsigned short)(y+1);
    u.ushort[5] = 3;
    if (ioctl(console_fd[vt], TIOCLINUX, &u.uchr[1]) < 0) {
      int my_errno = errno;
      if (my_errno == EIO && tried == 0) {
        reopen_console(vt);
        continue;               /* back to for */
      } else {
        syslog(LOG_ERR, "ioctl(TIOCLINUX) failed: %m");
        exc_eraise2(unix_error, my_errno);
      } /*if*/
    } /*if*/
    break;
  } /*for*/
}

void
disable_paste (int vt)
{
  set_bit(paste_disabled, vt);
}

void
paste_selection (int vt)
{
  static char arg = 3;
  int tried;

  if (test_bit(paste_disabled, vt)) {
    syslog(LOG_WARNING, "not pasting selection on tty%d, paste disabled", vt);
    return;
  } /*if*/

  if (max_selection_age >= 0 && time(0) > selection_time + max_selection_age) {
    syslog(LOG_WARNING, "not pasting selection on tty%d, selection too old", vt);
    return;
  } /*if*/

  do_open_console(vt);
  
  for (tried = 0; tried < 2; tried++) {
    if (ioctl(console_fd[vt], TIOCLINUX, &arg) < 0) {
      int my_errno = errno;
      if (my_errno == EIO && tried == 0) {
        reopen_console(vt);
        continue;               /* back to for */
      } else {
        syslog(LOG_ERR, "ioctl(TIOCLINUX) failed: %m");
        exc_eraise2(unix_error, my_errno);
      } /*if*/
    } /*if*/
    break;
  } /*for*/
}

void
copy_selection (int xs, int ys, int xe, int ye, int clicks, int vt)
{
  union {
    unsigned short ushort[6];
    unsigned char uchr[12];
  } u;
  int tried;

  do_open_console(vt);
  
  for (tried = 0; tried < 2; tried++) {
    u.uchr[1] = 2;
    u.ushort[1] = (unsigned short)(xs+1);
    u.ushort[2] = (unsigned short)(ys+1);
    u.ushort[3] = (unsigned short)(xe+1);
    u.ushort[4] = (unsigned short)(ye+1);
    u.ushort[5] = (unsigned short)(clicks-1);
    if (ioctl(console_fd[vt], TIOCLINUX, &u.uchr[1]) < 0) {
      int my_errno = errno;
      if (my_errno == EIO && tried == 0) {
        reopen_console(vt);
        continue;               /* back to for */
      } else {
        syslog(LOG_ERR, "ioctl(TIOCLINUX) failed: %m");
        exc_eraise2(unix_error, my_errno);
      } /*if*/
    } /*if*/
    break;
  } /*for*/

  selection_time = time(0);
  clear_bit(paste_disabled, vt);
}

void
load_inword (const unsigned long* inword_chars)
{
  union {
    unsigned long ulong[9];
    unsigned char uchr[36];
  } u;

  int tried;
  int i;
  const char* p;

  do_open_console(0);

  for (tried = 0; tried < 2; tried++) {
    u.uchr[0] = 5;
    for (i = 1; i < 9; i++) u.ulong[i] = inword_chars[i-1];
    if (0 > ioctl(console_fd[0], TIOCLINUX, &u.uchr[0])) {
      int my_errno = errno;
      if (tried == 0 && my_errno == EIO) {
        reopen_console(0);
        continue;               /* back to for */
      } else {
        syslog(LOG_ERR, "ioctl(TIOCLINUX) failed: %m");
        exc_eraise2(unix_error, my_errno);
      } /*if*/
    } /*if*/
    break;
  } /*for*/
}
